import ConnectorSingleton from './connector_singleton';
import { returnedSnippet } from '../typedefs';
import PiecesCacheSingleton from '../cache/pieces_cache';
import { AccessEnum, Asset, ConnectorSnapshot } from '../PiecesSDK/common';
import { Format } from '../PiecesSDK/common';
import { timeoutPromise } from '../utils';
import { mergeAssetsWithTransferables } from '../transferables';
import { ClassificationSpecificEnum } from '../PiecesSDK/connector';
import Notifications from './notification_handler';
import { setIsFetchFailed } from '../ui/views/create_snippet_view';

export const loadConnect = async (): Promise<boolean> => {
    const config: ConnectorSingleton = ConnectorSingleton.getInstance();
    try {
        config.context = await config.api.connect({
            seededConnectorConnection: config.seeded,
        });
        return true;
    } catch (err) {
        return false;
    }
};

export const loadPieces = async (): Promise<returnedSnippet[]> => {
    //
    const config: ConnectorSingleton = ConnectorSingleton.getInstance();
    const notifications: Notifications = Notifications.getInstance();

    if (!config.context) {
        try {
            config.context = await config.api.connect({
                seededConnectorConnection: config.seeded,
            });
        } catch (err) {
            // ignore we will just warn
        }

        if (!config.context) {
            notifications.error({
                message:
                    'Failed to connect to Pieces OS. Please check that Pieces OS is installed.',
            });
            setIsFetchFailed(true);
            return Promise.reject(new Error('Context Undefined'));
        }
    }

    const snapshot = await fetchSnapshot({ config: config });

    if (snapshot instanceof Error) {
        setIsFetchFailed(true);
        return Promise.reject(snapshot);
    }

    await fetchAllFormats({ assets: snapshot.assets.iterable });

    const piecesStorage = PiecesCacheSingleton.getInstance();
    const assets = mergeAssetsWithTransferables({
        assets: snapshot.assets,
        references: piecesStorage.formatTransferables,
    });

    /// (2) save our assets
    piecesStorage.store({ assets: assets.iterable });

    const { snippets } = await processAssets({
        assets: piecesStorage.assets,
    });

    // sort the snippets by date created (desc)
    setIsFetchFailed(false);
    //writeData(piecesStorage.assets);
    return snippets;
};

const fetchSnapshot = async ({
    config,
    retry = false,
}: {
    suggested?: boolean;
    config: ConnectorSingleton;
    retry?: boolean;
}): Promise<ConnectorSnapshot | Error> => {
    const notifications: Notifications = Notifications.getInstance();

    let snapshot: ConnectorSnapshot;

    try {
        snapshot = await config.api.snapshot({
            application: config.context.application.id,
            suggested: false,
            transferables: false,
        });
        return snapshot;
    } catch (error) {
        if (!retry) {
            return await fetchSnapshot({
                suggested: false,
                config: config,
                retry: true,
            });
        } else {
            notifications.error({
                message:
                    'Failed to load snippets. Please restart Pieces OS, ensure that it is up-to-date, and try again. If the problem persists please reach out to support at support@pieces.app.',
            });
            return Promise.reject('Failed to fetch snapshot');
        }
    }
};

// fetches all formats for the assets and stores them in cache
const fetchAllFormats = async ({ assets }: { assets: Asset[] }) => {
    const piecesStorage = PiecesCacheSingleton.getInstance();
    const formatFetching: Promise<void>[] = [];
    for (const asset of assets) {
        // TODO may need to tweak this value (every 10 format fetchs) as well as we may need to space our timeoutPromise a bit more, but shouldnt need too.
        // await timeoutPromise(0);

        /// if we want to fetch.
        for (const format of asset.formats.iterable || []) {
            /// (3) fetch any formats that we havnt fetched yet.
            if (
                !(format.id in piecesStorage.fetchedFormats) ||
                piecesStorage.fetchedFormats[format.id] < format.updated.value
            ) {
                /// if either (1) our format hasnt been fetched yet
                /// or if (2) our format was updated before the last time it has been fetched.
                /// then we will want to fetch our format.
                /// dont await this as it can just happen syncronously
                formatFetching.push(fetchFormatTransferable({ format, asset }));
            }
        }
    }
    await Promise.all(formatFetching);
};

/**
 * This is a helper function that will enable us to process our Assets and turn them into the ReturnedAssets type.
 * This will also return a list of promises, that will represent everything that is currently being fetched.
 * @param {Boolean} fetch if we want to also attempt to fetch the assets.
 * @param {Assets} assets
 * @returns
 */
export const processAssets = async ({ assets }: { assets: Asset[] }) => {
    const snippets: returnedSnippet[] = [];

    for (const asset of assets) {
        snippets.push(processAsset({ asset: asset }));
    }
    return {
        snippets,
    };
};

export const processAsset = ({ asset }: { asset: Asset }): returnedSnippet => {
    let rawText = undefined;
    let rawFormat = undefined;

    const type = asset.original.reference?.classification.generic;
    if (type === 'IMAGE') {
        const decoder = new TextDecoder('utf-8');
        const ocrid = asset.original.reference?.analysis?.image?.ocr?.raw.id;
        if (!ocrid) {
            // LOG IN SENTRY
        }
        const format = asset.formats.iterable?.find((e) => e.id === ocrid);
        const bytes = new Uint8Array(format?.file?.bytes?.raw ?? []);
        rawText = decoder.decode(bytes);
        rawFormat = format?.classification.specific;
    }

    let link: string | undefined = undefined;
    for (let i = 0; i < (asset.shares?.iterable.length ?? 0); i++) {
        if (asset.shares?.iterable[i].access === AccessEnum.Public) {
            link = asset.shares?.iterable[i].link;
            break;
        }
    }

    return {
        title: asset.name || 'Unnamed Asset',
        id: asset.id,
        type:
            asset.original.reference?.classification.generic || 'Unknown Type',
        raw:
            rawText ??
            asset.original.reference?.fragment?.string?.raw ??
            asset.preview.base.reference?.fragment?.string?.raw ??
            asset.original.reference?.file?.string?.raw ??
            asset.preview.base.reference?.file?.string?.raw ??
            'Unable to unpack snippet :(',
        //@ts-ignore
        language:
            rawFormat ??
            (asset.original.reference?.classification.specific ||
                ClassificationSpecificEnum.Ts),
        time: asset.created.readable || 'Unknown Time',
        created: asset.created.value,
        description: asset.description,
        updated: asset.updated.value,
        share: link ?? undefined,
    };
};

export const fetchFormatTransferable = async ({
    format,
    asset,
    retryCount = 0,
}: {
    format: Format;
    asset: Asset;
    retryCount?: number;
}): Promise<void> => {
    const connector: ConnectorSingleton = ConnectorSingleton.getInstance();
    const storage = PiecesCacheSingleton.getInstance();

    // This is going to syncronously pull down and update as a fetch finishes.
    return connector.formatApi
        .formatSnapshot({
            format: format.id,
            transferable: true,
        })
        .then((fetched: Format) => {
            storage.formatTransferables[format.id] = {
                file: fetched.file,
                fragment: fetched.fragment,
            };

            /// (3) format has been fetched (we will use this to determine on the next refresh, to see if we need to refrech the format or not.)
            storage.fetchedFormats[format.id] = new Date();
        })
        .catch((error) => {
            /// TODO: if it fails, we want to load it in when the user actually clicks on it and add it in the storage.
            /// We want to try atleast 5 times if we get the connect ECONNRESET error. We are looking for this error specifically because of the rapid fire of the requests to our server.
            if (
                (error.message ?? '').includes('connect ECONNRESET') &&
                retryCount++ <= 5
            ) {
                console.log(`Making request ${format.id} Retry: ${retryCount}`);
                timeoutPromise(200).then((_) => {
                    fetchFormatTransferable({
                        format,
                        asset,
                        retryCount: retryCount,
                    });
                });
            } else if (
                (error.message ?? '').includes('connect ECONNRESET') &&
                retryCount++ > 5
            ) {
                console.log(
                    `Max threshold hit for this format ${format.id} Error: ${error}`
                );
            }
        });
};
